@@ -60,6 +60,9 @@ module Distribution.Client.ProjectPlanning.Types (
     SetupScriptStyle(..),
   ) where
 
+import           Distribution.Client.Compat.Prelude
+import           Prelude ()
+
 import           Distribution.Client.TargetSelector
                    ( SubComponentTarget(..) )
 import           Distribution.Client.PackageHash
@@ -75,8 +78,7 @@ import           Distribution.Client.DistDirLayout
 import           Distribution.Backpack
 import           Distribution.Backpack.ModuleShape
 
-import           Distribution.Pretty
-import           Distribution.Verbosity
+import           Distribution.Verbosity (normal)
 import           Distribution.Types.ComponentRequestedSpec
 import           Distribution.Types.PkgconfigVersion
 import           Distribution.Types.PackageDescription (PackageDescription(..))
@@ -102,16 +104,9 @@ import           Distribution.Solver.Types.OptionalStanza
 import           Distribution.Compat.Graph (IsNode(..))
 import           Distribution.Simple.Utils (ordNub)
 
-import           Data.Map (Map)
 import qualified Data.Map as Map
-import           Data.Maybe (catMaybes)
-import           Data.Set (Set)
 import qualified Data.ByteString.Lazy as LBS
-import           Distribution.Compat.Binary
-import           GHC.Generics (Generic)
 import qualified Data.Monoid as Mon
-import           Data.Typeable
-import           Control.Monad
 import           System.FilePath ((</>))
 
 
@@ -158,6 +153,7 @@ data ElaboratedSharedConfig
   --TODO: [code cleanup] no Eq instance
 
 instance Binary ElaboratedSharedConfig
+instance Structured ElaboratedSharedConfig
 
 data ElaboratedConfiguredPackage
    = ElaboratedConfiguredPackage {
@@ -297,6 +293,8 @@ data ElaboratedConfiguredPackage
        elabTestFailWhenNoTestSuites :: Bool,
        elabTestTestOptions       :: [PathTemplate],
 
+       elabBenchmarkOptions      :: [PathTemplate],
+
        -- Setup.hs related things:
 
        -- | One of four modes for how we build and interact with the Setup.hs
@@ -307,6 +305,9 @@ data ElaboratedConfiguredPackage
        -- | The version of the Cabal command line interface that we are using
        -- for this package. This is typically the version of the Cabal lib
        -- that the Setup.hs is built against.
+       --
+       -- TODO: We might want to turn this into a enum,
+       -- yet different enum than 'CabalSpecVersion'.
        elabSetupScriptCliVersion :: Version,
 
        -- Build time related:
@@ -461,6 +462,7 @@ instance IsNode ElaboratedConfiguredPackage where
     nodeNeighbors = elabOrderDependencies
 
 instance Binary ElaboratedConfiguredPackage
+instance Structured ElaboratedConfiguredPackage
 
 data ElaboratedPackageOrComponent
     = ElabPackage   ElaboratedPackage
@@ -468,6 +470,7 @@ data ElaboratedPackageOrComponent
   deriving (Eq, Show, Generic)
 
 instance Binary ElaboratedPackageOrComponent
+instance Structured ElaboratedPackageOrComponent
 
 elabComponentName :: ElaboratedConfiguredPackage -> Maybe ComponentName
 elabComponentName elab =
@@ -659,6 +662,7 @@ data ElaboratedComponent
   deriving (Eq, Show, Generic)
 
 instance Binary ElaboratedComponent
+instance Structured ElaboratedComponent
 
 -- | See 'elabOrderDependencies'.
 compOrderDependencies :: ElaboratedComponent -> [UnitId]
@@ -707,6 +711,7 @@ data ElaboratedPackage
   deriving (Eq, Show, Generic)
 
 instance Binary ElaboratedPackage
+instance Structured ElaboratedPackage
 
 -- | See 'elabOrderDependencies'.  This gives the unflattened version,
 -- which can be useful in some circumstances.
@@ -738,6 +743,7 @@ data BuildStyle =
   deriving (Eq, Show, Generic)
 
 instance Binary BuildStyle
+instance Structured BuildStyle
 
 type CabalFileText = LBS.ByteString
 
@@ -755,6 +761,7 @@ data ComponentTarget = ComponentTarget ComponentName SubComponentTarget
   deriving (Eq, Ord, Show, Generic)
 
 instance Binary ComponentTarget
+instance Structured ComponentTarget
 
 -- | Unambiguously render a 'ComponentTarget', e.g., to pass
 -- to a Cabal Setup script.
@@ -830,3 +837,4 @@ data SetupScriptStyle = SetupCustomExplicitDeps
   deriving (Eq, Show, Generic, Typeable)
 
 instance Binary SetupScriptStyle
+instance Structured SetupScriptStyle
