@@ -1,5 +1,7 @@
 {-# LANGUAGE CPP, DeriveGeneric, DeriveFunctor,
              RecordWildCards, NamedFieldPuns #-}
+-- TODO
+{-# OPTIONS_GHC -fno-warn-incomplete-uni-patterns #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Client.TargetSelector
@@ -47,7 +49,6 @@ import Distribution.Types.UnqualComponentName
 import Distribution.Client.Types
          ( PackageLocation(..), PackageSpecifier(..) )
 
-import Distribution.Verbosity
 import Distribution.PackageDescription
          ( PackageDescription
          , Executable(..)
@@ -65,26 +66,19 @@ import Distribution.Simple.LocalBuildInfo
          , pkgComponents, componentName, componentBuildInfo )
 import Distribution.Types.ForeignLib
 
-import Distribution.Deprecated.Text
-         ( Text, display, simpleParse )
 import Distribution.Simple.Utils
          ( die', lowercase, ordNub )
 import Distribution.Client.Utils
          ( makeRelativeCanonical )
 
-import Data.Either
-         ( partitionEithers )
-import Data.Function
-         ( on )
 import Data.List
-         ( stripPrefix, partition, groupBy )
-import Data.Ord
-         ( comparing )
+         ( stripPrefix, groupBy )
+import qualified Data.List.NonEmpty as NE
 import qualified Data.Map.Lazy   as Map.Lazy
 import qualified Data.Map.Strict as Map
 import qualified Data.Set as Set
 import Control.Arrow ((&&&))
-import Control.Monad 
+import Control.Monad
   hiding ( mfilter )
 import qualified Distribution.Deprecated.ReadP as Parse
 import Distribution.Deprecated.ReadP
@@ -186,11 +180,13 @@ data SubComponentTarget =
      -- | A specific module within a component.
    | ModuleTarget ModuleName
 
-     -- | A specific file within a component.
+     -- | A specific file within a component. Note that this does not carry the
+     -- file extension.
    | FileTarget   FilePath
   deriving (Eq, Ord, Show, Generic)
 
 instance Binary SubComponentTarget
+instance Structured SubComponentTarget
 
 
 -- ------------------------------------------------------------
@@ -218,10 +214,10 @@ readTargetSelectorsWith :: (Applicative m, Monad m) => DirActions m
                         -> Maybe ComponentKindFilter
                         -> [String]
                         -> m (Either [TargetSelectorProblem] [TargetSelector])
-readTargetSelectorsWith dirActions@DirActions{..} pkgs mfilter targetStrs =
+readTargetSelectorsWith dirActions@DirActions{} pkgs mfilter targetStrs =
     case parseTargetStrings targetStrs of
       ([], usertargets) -> do
-        usertargets' <- mapM (getTargetStringFileStatus dirActions) usertargets
+        usertargets' <- traverse (getTargetStringFileStatus dirActions) usertargets
         knowntargets <- getKnownTargets dirActions pkgs
         case resolveTargetSelectors knowntargets usertargets' mfilter of
           ([], btargets) -> return (Right btargets)
@@ -433,6 +429,23 @@ forgetFileStatus t = case t of
     TargetStringFileStatus7 s1   s2 s3 s4
                                  s5 s6 s7 -> TargetString7 s1 s2 s3 s4 s5 s6 s7
 
+getFileStatus :: TargetStringFileStatus -> Maybe FileStatus
+getFileStatus (TargetStringFileStatus1 _ f)     = Just f
+getFileStatus (TargetStringFileStatus2 _ f _)   = Just f
+getFileStatus (TargetStringFileStatus3 _ f _ _) = Just f
+getFileStatus _                                 = Nothing
+
+setFileStatus :: FileStatus -> TargetStringFileStatus -> TargetStringFileStatus
+setFileStatus f (TargetStringFileStatus1 s1 _)       = TargetStringFileStatus1 s1 f
+setFileStatus f (TargetStringFileStatus2 s1 _ s2)    = TargetStringFileStatus2 s1 f s2
+setFileStatus f (TargetStringFileStatus3 s1 _ s2 s3) = TargetStringFileStatus3 s1 f s2 s3
+setFileStatus _ t                                    = t
+
+copyFileStatus :: TargetStringFileStatus -> TargetStringFileStatus -> TargetStringFileStatus
+copyFileStatus src dst =
+    case getFileStatus src of
+      Just f -> setFileStatus f dst
+      Nothing -> dst
 
 -- ------------------------------------------------------------
 -- * Resolving target strings to target selectors
@@ -502,9 +515,9 @@ resolveTargetSelector knowntargets@KnownTargets{..} mfilter targetStrStatus =
     projectIsEmpty = null knownPackagesAll
 
     classifyMatchErrors errs
-      | not (null expected)
-      = let (things, got:_) = unzip expected in
-        TargetSelectorExpected targetStr things got
+      | Just expectedNE <- NE.nonEmpty expected
+      = let (things, got:|_) = NE.unzip expectedNE in
+        TargetSelectorExpected targetStr (NE.toList things) got
 
       | not (null nosuch)
       = TargetSelectorNoSuch targetStr nosuch
@@ -581,7 +594,12 @@ data TargetSelectorProblem
    | TargetSelectorNoTargetsInProject
   deriving (Show, Eq)
 
-data QualLevel = QL1 | QL2 | QL3 | QLFull
+-- | Qualification levels.
+-- Given the filepath src/F, executable component A, and package foo:
+data QualLevel = QL1    -- ^ @src/F@
+               | QL2    -- ^ @foo:src/F | A:src/F@
+               | QL3    -- ^ @foo:A:src/F | exe:A:src/F@
+               | QLFull -- ^ @pkg:foo:exe:A:file:src/F@
   deriving (Eq, Enum, Show)
 
 disambiguateTargetSelectors
@@ -598,12 +616,19 @@ disambiguateTargetSelectors matcher matchInput exactMatch matchResults =
     -- So, here's the strategy. We take the original match results, and make a
     -- table of all their renderings at all qualification levels.
     -- Note there can be multiple renderings at each qualification level.
+
+    -- Note that renderTargetSelector won't immediately work on any file syntax
+    -- When rendering syntax, the FileStatus is always FileStatusNotExists,
+    -- which will never match on syntaxForm1File!
+    -- Because matchPackageDirectoryPrefix expects a FileStatusExistsFile.
+    -- So we need to copy over the file status from the input
+    -- TargetStringFileStatus, onto the new rendered TargetStringFileStatus
     matchResultsRenderings :: [(TargetSelector, [TargetStringFileStatus])]
     matchResultsRenderings =
       [ (matchResult, matchRenderings)
       | matchResult <- matchResults
       , let matchRenderings =
-              [ rendering
+              [ copyFileStatus matchInput rendering
               | ql <- [QL1 .. QLFull]
               , rendering <- renderTargetSelector ql matchResult ]
       ]
@@ -620,6 +645,8 @@ disambiguateTargetSelectors matcher matchInput exactMatch matchResults =
            then Map.insert matchInput (Match Exact 0 matchResults)
            else id)
       $ Map.Lazy.fromList
+        -- (matcher rendering) should *always* be a Match! Otherwise we will hit
+        -- the internal error later on.
           [ (rendering, matcher rendering)
           | rendering <- concatMap snd matchResultsRenderings ]
 
@@ -1033,7 +1060,7 @@ syntaxForm1File ps =
     (pkgfile, ~KnownPackage{pinfoId, pinfoComponents})
       -- always returns the KnownPackage case
       <- matchPackageDirectoryPrefix ps fstatus1
-    orNoThingIn "package" (display (packageName pinfoId)) $ do
+    orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
       (filepath, c) <- matchComponentFile pinfoComponents pkgfile
       return (TargetComponent pinfoId (cinfoName c) (FileTarget filepath))
   where
@@ -1131,7 +1158,7 @@ syntaxForm2PackageComponent ps =
     p <- matchPackage ps str1 fstatus1
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentName pinfoComponents str2
           return (TargetComponent pinfoId (cinfoName c) WholeComponent)
         --TODO: the error here ought to say there's no component by that name in
@@ -1143,7 +1170,7 @@ syntaxForm2PackageComponent ps =
     render (TargetComponent p c WholeComponent) =
       [TargetStringFileStatus2 (dispP p) noFileStatus (dispC p c)]
     render (TargetComponentUnknown pn (Left cn) WholeComponent) =
-      [TargetStringFileStatus2 (dispPN pn) noFileStatus (display cn)]
+      [TargetStringFileStatus2 (dispPN pn) noFileStatus (prettyShow cn)]
     render _ = []
 
 -- | Syntax: namespace : component
@@ -1176,7 +1203,7 @@ syntaxForm2PackageModule ps =
     p <- matchPackage ps str1 fstatus1
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           let ms = [ (m,c) | c <- pinfoComponents, m <- cinfoModules c ]
           (m,c) <- matchModuleNameAnd ms str2
           return (TargetComponent pinfoId (cinfoName c) (ModuleTarget m))
@@ -1222,7 +1249,7 @@ syntaxForm2PackageFile ps =
     p <- matchPackage ps str1 fstatus1
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           (filepath, c) <- matchComponentFile pinfoComponents str2
           return (TargetComponent pinfoId (cinfoName c) (FileTarget filepath))
       KnownPackageName pn ->
@@ -1321,7 +1348,7 @@ syntaxForm3PackageKindComponent ps =
     p <- matchPackage ps str1 fstatus1
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentKindAndName pinfoComponents ckind str3
           return (TargetComponent pinfoId (cinfoName c) WholeComponent)
       KnownPackageName pn ->
@@ -1349,7 +1376,7 @@ syntaxForm3PackageComponentModule ps =
     p <- matchPackage ps str1 fstatus1
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentName pinfoComponents str2
           orNoThingIn "component" (cinfoStrName c) $ do
             let ms = cinfoModules c
@@ -1401,7 +1428,7 @@ syntaxForm3PackageComponentFile ps =
     p <- matchPackage ps str1 fstatus1
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentName pinfoComponents str2
           orNoThingIn "component" (cinfoStrName c) $ do
             (filepath, _) <- matchComponentFile [c] str3
@@ -1492,7 +1519,7 @@ syntaxForm5MetaNamespacePackageKindComponent ps =
     p <- matchPackage  ps str3 noFileStatus
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentKindAndName pinfoComponents ckind str5
           return (TargetComponent pinfoId (cinfoName c) WholeComponent)
       KnownPackageName pn ->
@@ -1522,7 +1549,7 @@ syntaxForm7MetaNamespacePackageKindComponentNamespaceModule ps =
     p <- matchPackage  ps str3 noFileStatus
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentKindAndName pinfoComponents ckind str5
           orNoThingIn "component" (cinfoStrName c) $ do
             let ms = cinfoModules c
@@ -1560,7 +1587,7 @@ syntaxForm7MetaNamespacePackageKindComponentNamespaceFile ps =
     p <- matchPackage  ps str3 noFileStatus
     case p of
       KnownPackage{pinfoId, pinfoComponents} ->
-        orNoThingIn "package" (display (packageName pinfoId)) $ do
+        orNoThingIn "package" (prettyShow (packageName pinfoId)) $ do
           c <- matchComponentKindAndName pinfoComponents ckind str5
           orNoThingIn "component" (cinfoStrName c) $ do
             (filepath,_) <- matchComponentFile [c] str7
@@ -1644,10 +1671,10 @@ syntaxForm7 render f =
     match _ = mzero
 
 dispP :: Package p => p -> String
-dispP = display . packageName
+dispP = prettyShow . packageName
 
 dispPN :: PackageName -> String
-dispPN = display
+dispPN = prettyShow
 
 dispC :: PackageId -> ComponentName -> String
 dispC = componentStringName . packageName
@@ -1656,7 +1683,7 @@ dispC' :: PackageName -> ComponentName -> String
 dispC' = componentStringName
 
 dispCN :: UnqualComponentName -> String
-dispCN = display
+dispCN = prettyShow
 
 dispK :: ComponentKind -> String
 dispK = showComponentKindShort
@@ -1668,7 +1695,7 @@ dispF :: ComponentKind -> String
 dispF = showComponentKindFilterShort
 
 dispM :: ModuleName -> String
-dispM = display
+dispM = prettyShow
 
 
 -------------------------------
@@ -1723,7 +1750,7 @@ getKnownTargets :: (Applicative m, Monad m)
                 -> [PackageSpecifier (SourcePackage (PackageLocation a))]
                 -> m KnownTargets
 getKnownTargets dirActions@DirActions{..} pkgs = do
-    pinfo <- mapM (collectKnownPackageInfo dirActions) pkgs
+    pinfo <- traverse (collectKnownPackageInfo dirActions) pkgs
     cwd   <- getCurrentDirectory
     let (ppinfo, opinfo) = selectPrimaryPackage cwd pinfo
     return KnownTargets {
@@ -1754,8 +1781,8 @@ collectKnownPackageInfo _ (NamedPackage pkgname _props) =
     return (KnownPackageName pkgname)
 collectKnownPackageInfo dirActions@DirActions{..}
                   (SpecificSourcePackage SourcePackage {
-                    packageDescription = pkg,
-                    packageSource      = loc
+                    srcpkgDescription = pkg,
+                    srcpkgSource      = loc
                   }) = do
     (pkgdir, pkgfile) <-
       case loc of
@@ -1764,8 +1791,8 @@ collectKnownPackageInfo dirActions@DirActions{..}
           dirabs <- canonicalizePath dir
           dirrel <- makeRelativeToCwd dirActions dirabs
           --TODO: ought to get this earlier in project reading
-          let fileabs = dirabs </> display (packageName pkg) <.> "cabal"
-              filerel = dirrel </> display (packageName pkg) <.> "cabal"
+          let fileabs = dirabs </> prettyShow (packageName pkg) <.> "cabal"
+              filerel = dirrel </> prettyShow (packageName pkg) <.> "cabal"
           exists <- doesFileExist fileabs
           return ( Just (dirabs, dirrel)
                  , if exists then Just (fileabs, filerel) else Nothing
@@ -1799,7 +1826,7 @@ collectKnownComponentInfo pkg =
 
 
 componentStringName :: PackageName -> ComponentName -> ComponentStringName
-componentStringName pkgname (CLibName LMainLibName) = display pkgname
+componentStringName pkgname (CLibName LMainLibName) = prettyShow pkgname
 componentStringName _ (CLibName (LSubLibName name)) = unUnqualComponentName name
 componentStringName _ (CFLibName name)  = unUnqualComponentName name
 componentStringName _ (CExeName   name) = unUnqualComponentName name
@@ -1971,9 +1998,9 @@ matchPackageName :: [KnownPackage] -> String -> Match KnownPackage
 matchPackageName ps = \str -> do
     guard (validPackageName str)
     orNoSuchThing "package" str
-                  (map (display . knownPackageName) ps) $
+                  (map (prettyShow . knownPackageName) ps) $
       increaseConfidenceFor $
-        matchInexactly caseFold (display . knownPackageName) ps str
+        matchInexactly caseFold (prettyShow . knownPackageName) ps str
 
 
 matchPackageNameUnknown :: String -> Match KnownPackage
@@ -2056,7 +2083,7 @@ matchComponentKindAndName cs ckind str =
 
 guardModuleName :: String -> Match ()
 guardModuleName s =
-  case simpleParse s :: Maybe ModuleName of
+  case simpleParsec s :: Maybe ModuleName of
     Just _                   -> increaseConfidence
     _ | all validModuleChar s
         && not (null s)      -> return ()
@@ -2067,16 +2094,16 @@ guardModuleName s =
 
 matchModuleName :: [ModuleName] -> String -> Match ModuleName
 matchModuleName ms str =
-    orNoSuchThing "module" str (map display ms)
+    orNoSuchThing "module" str (map prettyShow ms)
   $ increaseConfidenceFor
-  $ matchInexactly caseFold display ms str
+  $ matchInexactly caseFold prettyShow ms str
 
 
 matchModuleNameAnd :: [(ModuleName, a)] -> String -> Match (ModuleName, a)
 matchModuleNameAnd ms str =
-    orNoSuchThing "module" str (map (display . fst) ms)
+    orNoSuchThing "module" str (map (prettyShow . fst) ms)
   $ increaseConfidenceFor
-  $ matchInexactly caseFold (display . fst) ms str
+  $ matchInexactly caseFold (prettyShow . fst) ms str
 
 
 matchModuleNameUnknown :: String -> Match ModuleName
@@ -2113,12 +2140,14 @@ matchComponentOtherFile :: [KnownComponent] -> String
                         -> Match (FilePath, KnownComponent)
 matchComponentOtherFile cs =
     matchFile
-      [ (file, c)
-      | c    <- cs
-      , file <- cinfoHsFiles c
-             ++ cinfoCFiles  c
-             ++ cinfoJsFiles c
+      [ (normalise (srcdir </> file), c)
+      | c      <- cs
+      , srcdir <- cinfoSrcDirs c
+      , file   <- cinfoHsFiles c
+               ++ cinfoCFiles  c
+               ++ cinfoJsFiles c
       ]
+      . normalise
 
 
 matchComponentModuleFile :: [KnownComponent] -> String
@@ -2130,7 +2159,8 @@ matchComponentModuleFile cs str = do
       , d <- cinfoSrcDirs c
       , m <- cinfoModules c
       ]
-      (dropExtension (normalise str))
+      (dropExtension (normalise str)) -- Drop the extension because FileTarget
+                                      -- is stored without the extension
 
 -- utils
 
@@ -2366,8 +2396,8 @@ matchInexactly cannonicalise key xs =
     -- the map of canonicalised keys to groups of inexact matches
     m' = Map.mapKeysWith (++) cannonicalise m
 
-matchParse :: Text a => String -> Match a
-matchParse = maybe mzero return . simpleParse
+matchParse :: Parsec a => String -> Match a
+matchParse = maybe mzero return . simpleParsec
 
 
 ------------------------------
